#!/usr/bin/env python3
"""
Windows Backup Manager (System Management Tool)

Parameters:
    --list: List backup configurations and history
    --create PATH: Create new backup of specified path
    --restore ID PATH: Restore backup to specified path
    --schedule INTERVAL: Schedule automatic backups
    --verify ID: Verify backup integrity
    --format FORMAT: Output format (table/json)

Requires: win32com, tabulate packages and admin privileges
Warning: Ensure sufficient disk space before creating backups
"""

import os
import sys
import json
import shutil
import hashlib
import logging
import datetime
import subprocess
from typing import List, Dict, Optional
from pathlib import Path

import schedule
import win32com.client
from tabulate import tabulate

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)

class BackupManager:
    """Manages file and system backups on Windows."""
    
    def __init__(self, backup_dir: str = None):
        """
        Initialize the BackupManager.
        
        Args:
            backup_dir: Directory to store backups (default: Documents/Backups)
        """
        if backup_dir:
            self.backup_dir = Path(backup_dir)
        else:
            docs_path = Path(os.path.expanduser("~/Documents"))
            self.backup_dir = docs_path / "Backups"
            
        self.backup_dir.mkdir(parents=True, exist_ok=True)
        self.wmi = win32com.client.GetObject("winmgmts:")
        
    def create_file_backup(self, source_path: str, name: str = None) -> Optional[str]:
        """
        Create a backup of a file or directory.
        
        Args:
            source_path: Path to file or directory to backup
            name: Optional name for the backup (default: auto-generated)
            
        Returns:
            Optional[str]: Path to backup file or None if failed
        """
        source = Path(source_path)
        if not source.exists():
            logger.error(f"Source path does not exist: {source_path}")
            return None
            
        try:
            # Generate backup name if not provided
            if not name:
                timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
                name = f"{source.name}_{timestamp}"
                
            # Create backup directory
            backup_path = self.backup_dir / name
            backup_path.mkdir(parents=True, exist_ok=True)
            
            # Copy files
            if source.is_file():
                shutil.copy2(source, backup_path)
            else:
                shutil.copytree(source, backup_path / source.name, 
                              dirs_exist_ok=True)
                
            # Create metadata file
            metadata = {
                "source": str(source),
                "timestamp": datetime.datetime.now().isoformat(),
                "type": "file",
                "checksum": self._calculate_checksum(source)
            }
            
            metadata_file = backup_path / "backup_metadata.json"
            with open(metadata_file, "w") as f:
                json.dump(metadata, f, indent=2)
                
            logger.info(f"Successfully created backup: {name}")
            return str(backup_path)
            
        except Exception as e:
            logger.error(f"Error creating backup: {e}")
            return None
    
    def create_system_backup(self, name: str = None) -> Optional[str]:
        """
        Create a system state backup using Windows built-in tools.
        
        Args:
            name: Optional name for the backup (default: auto-generated)
            
        Returns:
            Optional[str]: Path to backup or None if failed
        """
        try:
            # Generate backup name if not provided
            if not name:
                timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
                name = f"SystemBackup_{timestamp}"
                
            backup_path = self.backup_dir / name
            backup_path.mkdir(parents=True, exist_ok=True)
            
            # Create system state backup using wbadmin
            cmd = [
                "wbadmin", "start", "backup",
                "-backupTarget:", str(backup_path),
                "-include:C:",
                "-systemState",
                "-quiet"
            ]
            
            result = subprocess.run(cmd, capture_output=True, text=True)
            
            if result.returncode == 0:
                # Create metadata file
                metadata = {
                    "timestamp": datetime.datetime.now().isoformat(),
                    "type": "system",
                    "windows_version": self._get_windows_version()
                }
                
                metadata_file = backup_path / "backup_metadata.json"
                with open(metadata_file, "w") as f:
                    json.dump(metadata, f, indent=2)
                    
                logger.info(f"Successfully created system backup: {name}")
                return str(backup_path)
            else:
                logger.error(f"System backup failed: {result.stderr}")
                return None
                
        except Exception as e:
            logger.error(f"Error creating system backup: {e}")
            return None
    
    def restore_backup(self, backup_name: str, target_path: Optional[str] = None) -> bool:
        """
        Restore from a backup.
        
        Args:
            backup_name: Name of the backup to restore
            target_path: Optional target path for restoration
            
        Returns:
            bool: True if successful, False otherwise
        """
        backup_path = self.backup_dir / backup_name
        if not backup_path.exists():
            logger.error(f"Backup not found: {backup_name}")
            return False
            
        try:
            # Read metadata
            metadata_file = backup_path / "backup_metadata.json"
            with open(metadata_file) as f:
                metadata = json.load(f)
                
            if metadata["type"] == "file":
                source = Path(metadata["source"])
                target = Path(target_path) if target_path else source
                
                if backup_path.is_file():
                    shutil.copy2(backup_path, target)
                else:
                    shutil.copytree(backup_path, target, dirs_exist_ok=True)
                    
            else:  # System backup
                cmd = [
                    "wbadmin", "start", "recovery",
                    "-backupTarget:", str(backup_path),
                    "-machine:", "localhost",
                    "-quiet"
                ]
                
                result = subprocess.run(cmd, capture_output=True, text=True)
                if result.returncode != 0:
                    logger.error(f"System restore failed: {result.stderr}")
                    return False
                    
            logger.info(f"Successfully restored backup: {backup_name}")
            return True
            
        except Exception as e:
            logger.error(f"Error restoring backup: {e}")
            return False
    
    def list_backups(self) -> List[Dict]:
        """
        Get a list of available backups.
        
        Returns:
            List[Dict]: List of backups with their details
        """
        backups = []
        
        try:
            for backup_dir in self.backup_dir.iterdir():
                if backup_dir.is_dir():
                    metadata_file = backup_dir / "backup_metadata.json"
                    if metadata_file.exists():
                        with open(metadata_file) as f:
                            metadata = json.load(f)
                            
                        backups.append({
                            "name": backup_dir.name,
                            "type": metadata["type"],
                            "timestamp": metadata["timestamp"],
                            "size": self._get_dir_size(backup_dir)
                        })
                        
        except Exception as e:
            logger.error(f"Error listing backups: {e}")
            
        return backups
    
    def verify_backup(self, backup_name: str) -> bool:
        """
        Verify the integrity of a backup.
        
        Args:
            backup_name: Name of the backup to verify
            
        Returns:
            bool: True if backup is valid, False otherwise
        """
        backup_path = self.backup_dir / backup_name
        if not backup_path.exists():
            logger.error(f"Backup not found: {backup_name}")
            return False
            
        try:
            metadata_file = backup_path / "backup_metadata.json"
            with open(metadata_file) as f:
                metadata = json.load(f)
                
            if metadata["type"] == "file":
                source = Path(metadata["source"])
                if source.exists():
                    original_checksum = metadata["checksum"]
                    current_checksum = self._calculate_checksum(source)
                    
                    if original_checksum == current_checksum:
                        logger.info(f"Backup verified successfully: {backup_name}")
                        return True
                    else:
                        logger.warning(f"Backup checksum mismatch: {backup_name}")
                        return False
                else:
                    logger.warning(f"Original source not found: {source}")
                    return False
            else:
                # For system backups, use wbadmin to verify
                cmd = [
                    "wbadmin", "start", "verifybackup",
                    "-backupTarget:", str(backup_path),
                    "-quiet"
                ]
                
                result = subprocess.run(cmd, capture_output=True, text=True)
                if result.returncode == 0:
                    logger.info(f"System backup verified successfully: {backup_name}")
                    return True
                else:
                    logger.warning(f"System backup verification failed: {backup_name}")
                    return False
                    
        except Exception as e:
            logger.error(f"Error verifying backup: {e}")
            return False
    
    def _calculate_checksum(self, path: Path) -> str:
        """
        Calculate SHA-256 checksum of a file or directory.
        
        Args:
            path: Path to calculate checksum for
            
        Returns:
            str: Hexadecimal checksum
        """
        hasher = hashlib.sha256()
        
        if path.is_file():
            with open(path, "rb") as f:
                for chunk in iter(lambda: f.read(4096), b""):
                    hasher.update(chunk)
        else:
            for subpath in sorted(path.rglob("*")):
                if subpath.is_file():
                    hasher.update(subpath.relative_to(path).as_posix().encode())
                    with open(subpath, "rb") as f:
                        for chunk in iter(lambda: f.read(4096), b""):
                            hasher.update(chunk)
                            
        return hasher.hexdigest()
    
    def _get_dir_size(self, path: Path) -> int:
        """
        Get the total size of a directory in bytes.
        
        Args:
            path: Directory path
            
        Returns:
            int: Total size in bytes
        """
        return sum(f.stat().st_size for f in path.rglob("*") if f.is_file())
    
    def _get_windows_version(self) -> str:
        """
        Get the Windows version information.
        
        Returns:
            str: Windows version string
        """
        try:
            os_info = self.wmi.ExecQuery("SELECT * FROM Win32_OperatingSystem")[0]
            return f"{os_info.Caption} {os_info.Version}"
        except Exception:
            return "Unknown Windows Version"

def main():
    """Main function to handle command-line interface."""
    import argparse
    
    parser = argparse.ArgumentParser(description="Windows Backup Manager")
    parser.add_argument("--backup-dir", help="Directory to store backups")
    parser.add_argument("--backup-file", help="Create backup of file/directory")
    parser.add_argument("--backup-system", action="store_true",
                      help="Create system state backup")
    parser.add_argument("--restore", help="Restore specified backup")
    parser.add_argument("--target", help="Target path for restoration")
    parser.add_argument("--list", action="store_true",
                      help="List available backups")
    parser.add_argument("--verify", help="Verify specified backup")
    parser.add_argument("--format", choices=["table", "json"], default="table",
                      help="Output format (table or json)")
    
    args = parser.parse_args()
    
    backup_manager = BackupManager(args.backup_dir)
    
    try:
        if args.backup_file:
            result = backup_manager.create_file_backup(args.backup_file)
            if result:
                print(f"Backup created successfully: {result}")
                
        elif args.backup_system:
            result = backup_manager.create_system_backup()
            if result:
                print(f"System backup created successfully: {result}")
                
        elif args.restore:
            if backup_manager.restore_backup(args.restore, args.target):
                print(f"Successfully restored backup: {args.restore}")
                
        elif args.list:
            backups = backup_manager.list_backups()
            if args.format == "json":
                print(json.dumps(backups, indent=2))
            else:
                headers = ["Name", "Type", "Timestamp", "Size (bytes)"]
                table_data = [[b["name"], b["type"], b["timestamp"], 
                             b["size"]] for b in backups]
                print(tabulate(table_data, headers=headers, tablefmt="grid"))
                
        elif args.verify:
            if backup_manager.verify_backup(args.verify):
                print(f"Backup verified successfully: {args.verify}")
            else:
                print(f"Backup verification failed: {args.verify}")
                
        else:
            parser.print_help()
            
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        sys.exit(1)

if __name__ == "__main__":
    main()
